package jam;

import java.awt.*;
import java.applet.*;
import java.net.URL;
import java.util.*;

/*************************************************************************
**
**    This class is an extension of Graph2D class.
**    It adds interactive selection of the plotting range
**    and can display the mouse position in user coordinates.
**
**    MouseDown: Starts the range selection
**    MouseDrag: Drag out a rectangular range selection
**    MouseUp:   Replot with modified plotting range.
**
**    Shift MouseUp: reset back to default range.
**
**    Cntrl MouseUp: Popup a window displaying the current mouse position
**    Cntrl MouseUp: Hide the position window.
**
*************************************************************************/


class G2Dint extends Graph2D {

/*
**    Set to true when a rectangle is being dragged out by the mouse
*/
      private boolean drag = false;

/*
**    Ths popup window for the cursor position command
*/
      private Gin cpgin = new Gin("Position");
/*
**    Ths popup window for the data point command
*/
      private Gin dpgin = new Gin("Data Point");
/*
**    Button Down position
*/
      private int x0,y0;
/*
**    Button Drag position
*/
      private int  x1,y1;
/*
**    Previous Button Drag position
*/
      private int x1old, y1old;

/*
**    Attached Axis which must be registered with this class
**    These are the axis used to find the drag range.
**    If no axis are registered no mouse drag.
*/
      private Axis xaxis;
      private Axis yaxis;


/*
**    Create xaxis to be used for the drag scaling
*/
      public Axis createXAxis() {
         xaxis = super.createAxis(Axis.BOTTOM);
         return xaxis;
      }
/*
**    Create yaxis to be used for the drag scaling
*/
      public Axis createYAxis() {
         yaxis = super.createAxis(Axis.LEFT);
         return yaxis;
      }
/*
**    Attach axis to be used for the drag scaling
*/
      public void attachAxis(Axis a) {
         if(a==null) return;

         super.attachAxis(a);         

         if(a.getAxisPos() == Axis.BOTTOM || a.getAxisPos() == Axis.TOP) {
              xaxis = a;
         } else {
              yaxis = a;
         }
      }
/*
**  Override the normal Graph2D update method with this new one
*/
    public void update(Graphics g) {
          Rectangle r = bounds();
          Color c = g.getColor();

          if(drag) {
/*
**         Drag out the new box.
**         Use drawLine instead of drawRect to avoid problems
**         when width and heights become negative. Seems drawRect
**         can't handle it!
*/

           g.setColor(getBackground());
           g.drawLine(x0, y0, x1old, y0);
           g.drawLine(x1old, y0, x1old, y1old);
           g.drawLine(x1old, y1old, x0, y1old);
           g.drawLine(x0, y1old, x0, y0);
           g.setColor(getForeground());
           g.drawLine(x0, y0, x1, y0);
           g.drawLine(x1, y0, x1, y1);
           g.drawLine(x1, y1, x0, y1);
           g.drawLine(x0, y1, x0, y0);


           g.setColor(c);

           x1old = x1;
           y1old = y1;

           return;
           }

          if( clearAll ) {
             g.setColor(getBackground());
             g.fillRect(r.x,r.y,r.width,r.height);
             g.setColor(c);
          }
          if( paintAll ) paint(g);
    }

/*
** Handle the events
*/
    public boolean handleEvent(Event e) {

        if( xaxis == null || yaxis == null ) return false;

        switch (e.id) {
	  case Event.KEY_PRESS:
               switch ( e.key ) {

               case 'R':
                          xaxis.resetRange();
                          yaxis.resetRange();

                          repaint();
                          return true;
               case 'r':
                              repaint();
                              return true;
 	       case 'c':
                             if( cpgin.isVisible() ) {
                                     cpgin.hide();
  		             } else {
                                    cpgin.show();
		             }
                             return true;
 	       case 'D':
                            dpgin.hide();
                            return true; 
 	       case 'd':
                            dpgin.show();
                            double d[] = getClosestPoint(e.x, e.y);
                            dpgin.setXlabel( d[0] );
                            dpgin.setYlabel( d[1] );
                            if( xaxis != null && yaxis != null ) {
                                   int ix = xaxis.getInteger(d[0]);
                                   int iy = yaxis.getInteger(d[1]);
                                   if( ix >= datarect.x &&
                                       ix <= datarect.x +datarect.width && 
                                       iy >= datarect.y &&
                                       iy <= datarect.y +datarect.height ) {
                                       Graphics g = getGraphics();
                                       g.fillOval(ix-4, iy-4, 8, 8);
				     }
 			    }
                            return true;
	       default:
//                             System.out.println("KeyPress "+e.key);
                             return false;
               }
           case Event.MOUSE_DOWN:

                x0 = e.x;
                y0 = e.y;

                drag = true; 
                x1old = x0;
                y1old = y0;


                if(x0 < datarect.x) x0 = datarect.x;
                else
                if(x0 > datarect.x + datarect.width ) 
                    x0 = datarect.x + datarect.width;

                if(y0 < datarect.y) y0 = datarect.y;
                else
                if(y0 > datarect.y + datarect.height ) 
                    y0 = datarect.y + datarect.height;


                return true;
          case Event.MOUSE_UP:
                x1   = e.x;
                y1   = e.y;

                drag = false;


                if(x1 < datarect.x) x1 = datarect.x;
                else
                if(x1 > datarect.x + datarect.width ) 
                    x1 = datarect.x + datarect.width;

                if(y1 < datarect.y) y1 = datarect.y;
                else
                if(y1 > datarect.y + datarect.height ) 
                    y1 = datarect.y + datarect.height;


                if( Math.abs(x0-x1) > 5 &&  Math.abs(y0-y1) > 5 ) {
                   if(x0 < x1 ) {                
                      xaxis.minimum = xaxis.getDouble(x0);
                      xaxis.maximum = xaxis.getDouble(x1);
                   } else {
                      xaxis.maximum = xaxis.getDouble(x0);
                      xaxis.minimum = xaxis.getDouble(x1);
                   }

                   if(y0 >y1 ) {                
                      yaxis.minimum = yaxis.getDouble(y0);
                      yaxis.maximum = yaxis.getDouble(y1);
                   } else {
                      yaxis.maximum = yaxis.getDouble(y0);
                      yaxis.minimum = yaxis.getDouble(y1);
                   }

                   repaint();
                 }
                return true;

           case Event.MOUSE_DRAG:
                

                x1   = e.x;
                y1   = e.y;

                if(drag) {

                  if(x1 < datarect.x) x1 = datarect.x;
                  else
                  if(x1 > datarect.x + datarect.width ) 
                     x1 = datarect.x + datarect.width;

                  if(y1 < datarect.y) y1 = datarect.y;
                  else
                  if(y1 > datarect.y + datarect.height ) 
                     y1 = datarect.y + datarect.height;

                }
                
                if(cpgin.isVisible()) {
                            cpgin.setXlabel(  xaxis.getDouble(x1) );
                            cpgin.setYlabel(  yaxis.getDouble(y1) );
                }

                repaint();

                return true;

           case Event.MOUSE_MOVE:
                

                x1   = e.x;
                y1   = e.y;

                
                if(cpgin.isVisible()) {
                            cpgin.setXlabel(  xaxis.getDouble(x1) );
                            cpgin.setYlabel(  yaxis.getDouble(y1) );
                }

                return true;
           }

           return false;
       
         }

/*
**   Find the closest data point to the cursor
*/
     protected double[] getClosestPoint(int cx, int cy) {
        Enumeration e;
        DataSet ds;
        int   i;
        double a[] = new double[3];
        double dist2 = -1.0;
        double data[] = {0.0, 0.0};

        for ( e = dataset.elements() ; e.hasMoreElements() ;) {
               ds = ((DataSet)e.nextElement());

               a = ds.getClosestPoint(cx,cy);

               if( dist2 < 0.0 || dist2 > a[3] ) {
                                    data[0] = a[0];
                                    data[1] = a[1];
	        }
	}
        return data;

      }

}

/*
** Gin
**      Popup a window to output data after a Graphics Input command
*/

class Gin extends Frame {

     private Label xlabel = new Label();
     private Label ylabel = new Label();


     public Gin() {

         setLayout(new GridLayout(2,1) ) ;

         xlabel.setAlignment(Label.LEFT);
         ylabel.setAlignment(Label.LEFT);

         this.setFont(new Font("Helvetica", Font.PLAIN, 20));

         add("x",xlabel);
         add("y",ylabel);

         resize(110,80);
         super.setTitle("Graphics Input");
     }

     public Gin(String title) {
         this();
         if(title != null) super.setTitle(title);
       }

     public void setXlabel(double d) {
         xlabel.setText( String.valueOf(d) );
     }

     public void setYlabel(double d) {
         ylabel.setText( String.valueOf(d) );
     }

     public void setLabels(double dx, double dy) {
         xlabel.setText( String.valueOf(dx) );
         ylabel.setText( String.valueOf(dy) );
     }

     public void setFont( Font f ) {

        if ( f == null ) return;
        xlabel.setFont( f );
        ylabel.setFont( f );

     }     

     public void resize( int x, int y) {
        super.resize(x,y);
     }

}



