package jam;

import java.awt.*;
import java.lang.*;
import java.io.*;


public class RuntimeHandler extends Thread
{
          
 int layer_typ;		// The layer concerned

 int con_typ;		// The typ of data shown ( video, audio )

 Packet packets[];	// The packet array to hold the actual data packets

 TempFileHandler tfh;	// The concerned TempFileHandler to deliver packets

 ConferenceMenu menu;	// The calling ConferenceMenu

 int packet_counter;	// The counter to avoid array overflow

 String tmp_file;       // The file containing the temporary data
  
 String end_file;	// The file(s) where the DataSet's are written 
			// to at the end of a conference

 // The different GraphHandlers for visualization

 GraphHandler grh_iat = null;	// For showing the interarrival time
 GraphHandler grh_bw  = null;	// For showing the bandwidth usage
 GraphHandler grh_lr  = null;	// For showing the loss-rate


 // The DataSets for visualization

 DataSet ds_iat;
 DataSet ds_bw;
 DataSet ds_lr;


 // Some help variables 
 
 int tmp_count = 0;
 double sum_time = 0;
 double sum_len  = 0;
 int last_seq_no = -1;

 // The start_time and start_bytes for RTCP - packets
 
 double start_time, start_bytes, last_bytes;



 public RuntimeHandler( int lay, int data, ConferenceMenu parent )
 {

  layer_typ = lay;
  con_typ = data;
  menu = parent;
  packet_counter = 0;

  setPriority(Thread.MIN_PRIORITY);

  tmp_file = Config.HOME + menu.get_conference().get_name();


  switch ( layer_typ ) {

   case Config.LAYER_ATM : 

           // nothing to do right now;
          break;

   case Config.LAYER_IP  : 

           ds_iat = new DataSet();
           grh_iat = new GraphHandler(ds_iat, Config.GRAPH_INTERARRIVAL_TIME,
                                      con_typ, layer_typ);
           ds_bw = new DataSet();
           grh_bw = new GraphHandler(ds_bw,Config.GRAPH_BANDWIDTH_USAGE,
				con_typ, layer_typ);
           ds_lr = new DataSet();
           grh_lr = new GraphHandler(ds_lr, Config.GRAPH_LOSSRATE, con_typ, 
				layer_typ);
           packet_counter = Config.MAX_TMP_IP_PACKETS;
           packets = new IP_Packet[packet_counter];
           tmp_file = tmp_file + ".ip";
           break;

   case Config.LAYER_UDP :

           ds_iat = new DataSet();
           grh_iat = new GraphHandler(ds_iat, Config.GRAPH_INTERARRIVAL_TIME,
                                      con_typ, layer_typ);
           ds_bw = new DataSet();
           grh_bw = new GraphHandler(ds_bw,Config.GRAPH_BANDWIDTH_USAGE, 
				con_typ, layer_typ);
           packet_counter = Config.MAX_TMP_UDP_PACKETS;
           packets = new UDP_Packet[packet_counter];
           tmp_file = tmp_file + ".udp.";
           if ( con_typ == Config.AUDIO_CONNECTION ) { 
            tmp_file = tmp_file + Config.AUDIO_STR; }
           else tmp_file = tmp_file + Config.VIDEO_STR;
           break;

   case Config.LAYER_TCP : 

           // How the hell is the tcp-output of snoop
           packet_counter = Config.MAX_TMP_UDP_PACKETS;    
           packets = new TCP_Packet[packet_counter];
           break;

   case Config.LAYER_RTP :

           ds_bw = new DataSet();
           grh_bw = new GraphHandler(ds_bw, Config.GRAPH_BANDWIDTH_USAGE, 
				con_typ, layer_typ);
           ds_lr = new DataSet();
           grh_lr = new GraphHandler(ds_lr, Config.GRAPH_LOSSRATE, con_typ, 
				layer_typ);
           packet_counter = Config.MAX_TMP_RTCP_PACKETS;
           packets = new RTCP_Packet[packet_counter];
           tmp_file = tmp_file + ".rtp.";  
           if ( con_typ == Config.AUDIO_CONNECTION ) { 
            tmp_file = tmp_file + Config.AUDIO_STR; }
           else tmp_file = tmp_file + Config.VIDEO_STR;
           break;
  }

  // Reson for this sleep command is that RH tries to read the
  // file before snoop/getrtcp create the file. 
  try { sleep(1000); } catch ( Exception e ) {}

  tfh = new TempFileHandler( this );
  end_file = tmp_file;

 }


 public void accept_packet( Packet p )
 {
   // Check which Packet arrived and put the new packet into the
   // temporary list of packets

   switch ( layer_typ ) {

     case Config.LAYER_ATM :

                 ATM_Packet atm = (ATM_Packet)p;
                 try { packets[tmp_count++] = atm; } catch (Exception e ) {}
                 break;

     case Config.LAYER_IP  :

                 IP_Packet ip =  (IP_Packet)p    ;
                 try { packets[tmp_count++] = ip; } catch (Exception e ) {}
                 break;

     case Config.LAYER_UDP :

                 UDP_Packet udp = (UDP_Packet)p;
                 try { packets[tmp_count++] = udp; } catch (Exception e ) {}
                 break;

     case Config.LAYER_TCP :

                 TCP_Packet tcp = (TCP_Packet)p;
                 try { packets[tmp_count++] = tcp; } catch (Exception e ) {}
                 break;

     case Config.LAYER_RTP :

                 RTCP_Packet rtcp = (RTCP_Packet)p;
                 try { packets[tmp_count++] = rtcp; } catch (Exception e ) {}
                 break;
 
   }
 }


 public void update_datasets()
 {
   if ( tmp_count >  0 ) {

//    System.out.println("RuntimeHandler : rec. " + 
//                tmp_count + " packet(s) at layer " + layer_typ);
     // Update the bandwidth usage graphics
     update_bandwidth_usage_dataset();
  
     // Update the lossrate graphics
     update_lossrate_dataset();
  
     // update interarrival jitter
     update_interarrival_time_dataset();

   }
 }


 // Update the bandwidth usage graphics
 public void update_bandwidth_usage_dataset()
 {
   // Sum up the total time and amount of bytes received
   double bw, dbl[];

   switch ( layer_typ ) {

    case Config.LAYER_IP : 

      for ( int i = 0; i < tmp_count; i++ ) {
        IP_Packet ip = (IP_Packet)packets[i];
        sum_time += ip.time;
        sum_len  += ip.len;
      }

      // Compute the bandwidth for this period
      bw = ((( sum_len * 8 ) / 1000) / sum_time);
      dbl = new double[2];
      dbl[0] = ds_bw.getXmax() + sum_time;
      dbl[1] = bw;
      try { ds_bw.append( dbl, 1 ); } catch ( Exception e ) {}
      break;


     case Config.LAYER_UDP :

      for ( int i = 0; i < tmp_count; i++ ) {
        UDP_Packet udp = (UDP_Packet)packets[i];
        sum_time += udp.time;
        sum_len  += udp.len;
      }
 
      // Compute the bandwidth for this period
      bw = ((( sum_len * 8 ) / 1000) / sum_time);
      dbl = new double[2];
      dbl[0] = ds_bw.getXmax() + sum_time;
      dbl[1] = bw;
      try { ds_bw.append( dbl, 1 ); } catch ( Exception e ) {}
      break;


    case Config.LAYER_RTP :

      // Log each RTCP packet
      RTCP_Packet rtcp = (RTCP_Packet)packets[0];
      if ( ds_bw.dataPoints() == 0) {
         start_time = rtcp.time;
         start_bytes = rtcp.bytes;
         last_bytes = start_bytes;
       }
      bw = 0;
      sum_time = rtcp.time - start_time;

      // To avoid division by zero
      if ( ds_bw.dataPoints() != 0)
       bw = ( ( ( ( rtcp.bytes - last_bytes ) * 8 ) / 1000 ) 
                / 
                ( rtcp.time - (start_time + ds_bw.getXmax() )));
      last_bytes = rtcp.bytes;
      dbl = new double[2];
      dbl[0] = sum_time;
      if ( bw < 0 ) bw = 0;
      dbl[1] = bw;
      try { ds_bw.append( dbl, 1 ); } catch ( Exception e ) {}
      break;
   }
 }



 // Update the lossrate graphics
 public void  update_lossrate_dataset()
 {
   int lost_number = 0;
   int diff, seq_j, seq_j_1;
   double dbl[];

   switch ( layer_typ ) {

    case Config.LAYER_IP :

          IP_Packet ip;

          if ( last_seq_no == -1 ) { 
            ip = (IP_Packet)packets[0]; 
            last_seq_no = ip.seq;
          }
          for ( int j = 0; j < tmp_count; j++ ) {
            ip = (IP_Packet)packets[j];
            seq_j = ip.seq;
            if ( j == 0 )
              seq_j_1 = last_seq_no;
            else {
              ip = (IP_Packet)packets[j-1];
              seq_j_1 = ip.seq;
            }
            if ( seq_j < seq_j_1 ) /* overflow */ {
            } 
            else {
             lost_number += ( seq_j - seq_j_1 - 1 ); 
            }
          }

          ip = (IP_Packet)packets[tmp_count-1];
          last_seq_no = ip.seq;
          double loss_rate = (100.0 * (double)lost_number) 
                              / ((double)lost_number + (double)tmp_count);
          if ( loss_rate < 0 ) loss_rate = 0;
          dbl = new double[2];
          // sum_time is taken from the bandwidth-usage  method
          dbl[0] = ds_bw.getXmax() + sum_time;
          dbl[1] = loss_rate;
          try { ds_lr.append( dbl, 1 ); } catch ( Exception e ) {}
          break;



     case Config.LAYER_RTP :

          RTCP_Packet rtcp = (RTCP_Packet)packets[0];
          dbl = new double[2];
          dbl[0] = sum_time;
          dbl[1] = rtcp.lost; 
          try { ds_lr.append( dbl, 1 ); } catch ( Exception e ) {}
          break;
   }
 }

 
 // update interarrival jitter
 public void update_interarrival_time_dataset()
 {
   double dbl[];

   // Add all items from the temp-list to the DataSet

   switch ( layer_typ ) {

     case Config.LAYER_IP :

       for ( int i = 0; i < tmp_count; i++ ) {

          dbl = new double[2];
          IP_Packet ip = (IP_Packet)packets[i];
          dbl[0] = ip.time;
          dbl[1] = 1;

          try { ds_iat.append( dbl, 1 ); } catch ( Exception e ) {} 
       }
       ds_iat.compress();
       break;


      case Config.LAYER_UDP :

       for ( int i = 0; i < tmp_count; i++ ) {

          dbl = new double[2];
          UDP_Packet udp = (UDP_Packet)packets[i];
          dbl[0] = udp.time;
          dbl[1] = 1;

          try { ds_iat.append( dbl, 1 ); } catch ( Exception e ) {} 
       }
       ds_iat.compress();
       break;

   }
 }
 



 public void run()
 {
   while ( Config.CONFERENCE_IS_RUNNING == true ) {

    // Call TempFileHandler for new data
    tfh.readfile();

    // Update all DataSets for GraphHandlers
    update_datasets();

    // Call GraphHandlers to repaint graphs

    if ( grh_iat != null ) grh_iat.actualize();
    if ( grh_bw  != null ) grh_bw.actualize();
    if ( grh_lr  != null ) grh_lr.actualize();

    // Reset the counters

    tmp_count = 0;
    sum_time = 0;
    sum_len  = 0;

    // Go sleep

    try { 
      switch ( layer_typ ) {
       case Config.LAYER_ATM : sleep ( Config.CYCLETIME_ATM ); break;
       case Config.LAYER_IP  : sleep ( Config.CYCLETIME_IP  ); break;
       case Config.LAYER_UDP : sleep ( Config.CYCLETIME_UDP ); break;
       case Config.LAYER_TCP : sleep ( Config.CYCLETIME_TCP ); break;
       case Config.LAYER_RTP : sleep ( Config.CYCLETIME_RTP ); break;
      }
    } catch ( Exception e ) {}
   }
 }


 public void showgraph( int typ ) {

  switch ( typ ) {

   case Config.GRAPH_INTERARRIVAL_TIME :
          if ( grh_iat  != null ) {
           grh_iat.show();
           grh_iat.actualize();
          }
          break;

   case Config.GRAPH_LOSSRATE :
          if ( grh_lr != null ) {
           grh_lr.show();
           grh_lr.actualize();
          }
          break;

   case Config.GRAPH_BANDWIDTH_USAGE :
          if ( grh_bw != null ) {
           grh_bw.show();
           grh_bw.actualize();
          }
          break;
  }
 }

 public void terminate() 
 {
  String graph_typ_str = "#GRAPH_TYP\t";
  byte b[] = new byte[30];
  FileOutputStream fout_iat, fout_bw, fout_lr;
  String s;

  try {

    if ( grh_iat != null ) {

      fout_iat = new FileOutputStream(end_file + ".iat");
      graph_typ_str.getBytes( 0, graph_typ_str.length(), b, 0);
      fout_iat.write(b,0, graph_typ_str.length());
      s = Integer.toString( Config.GRAPH_INTERARRIVAL_TIME );
      s.getBytes( 0, s.length(), b, 0);
      fout_iat.write(b, 0, s.length());
      fout_iat.write('\n');
      writeHeader( fout_iat );
      ds_iat.compress();
      writeDataSet( fout_iat, ds_iat );
      fout_iat.close();
      grh_iat.hide();
    }

    if ( grh_lr != null ) {

      fout_lr = new FileOutputStream(end_file + ".lr");
      graph_typ_str.getBytes( 0, graph_typ_str.length(), b, 0);
      fout_lr.write(b,0, graph_typ_str.length());      
      s = Integer.toString( Config.GRAPH_LOSSRATE);
      s.getBytes( 0, s.length(), b, 0);
      fout_lr.write(b, 0, s.length());
      fout_lr.write('\n');
      writeHeader( fout_lr );
      writeDataSet( fout_lr, ds_lr );
      fout_lr.close();
      grh_lr.hide();
    }

    if ( grh_bw != null ) {

      fout_bw = new FileOutputStream(end_file + ".bw");
      graph_typ_str.getBytes( 0, graph_typ_str.length(), b, 0);
      fout_bw.write(b,0, graph_typ_str.length());
      s = Integer.toString( Config.GRAPH_BANDWIDTH_USAGE);
      s.getBytes( 0, s.length(), b, 0);
      fout_bw.write(b, 0, s.length());
      fout_bw.write('\n');
      writeHeader( fout_bw );
      writeDataSet( fout_bw, ds_bw );
      fout_bw.close();
      grh_bw.hide();
    }
  } catch ( Exception e ) { System.out.println("Error when writing to file " +
                                                end_file);}
  tfh.delete_file();
  this.stop();
 }


 public void writeHeader( FileOutputStream r ) {

  String con_typ_str = "#CONNECTION_TYP\t";
  String lay_typ_str = "#LAYER_TYP\t";
  String s;
  byte b[] = new byte[30];

  try {
   con_typ_str.getBytes( 0, con_typ_str.length(), b, 0);
   r.write(b, 0, con_typ_str.length());
   s = Integer.toString(con_typ);
   s.getBytes( 0, s.length(), b, 0);
   r.write(b, 0, s.length());
   r.write('\n');
   lay_typ_str.getBytes( 0, lay_typ_str.length(), b, 0);
   r.write(b, 0, lay_typ_str.length());
   s = Integer.toString( layer_typ );
   s.getBytes( 0, s.length(), b, 0);
   r.write(b, 0, s.length());
   r.write('\n');
  } catch ( IOException e) { System.out.println("Error when writing to file " +
                                                end_file); }
 }


 public void writeDataSet( FileOutputStream r, DataSet ds ) {

 // Now write the reamining DataSet
 String s;
 byte b[] = new byte[30];

 try {

  for ( int i = 0; i < ds.dataPoints(); i++ ) {

   double dbl[] = ds.getPoint(i);
   s = Double.toString( dbl[0] );
   s.getBytes(0, s.length(), b, 0 );
   r.write(b, 0, s.length());
   r.write('\t');
   s = Double.toString( dbl[1] );
   s.getBytes(0, s.length(), b, 0 );
   r.write(b, 0, s.length());
   r.write('\n');

  }

 } catch ( IOException e) { System.out.println("Error when writing to file " +
                                                end_file);}
 }

}